/*
 * FindBugs - Find bugs in Java programs
 *
 * (C) Copyright 2005 Diomidis Spinellis <dds@aueb.gr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: StringV.java,v 1.1.1.1 2005/07/29 09:20:43 dds Exp $
 */

package edu.umd.cs.findbugs.apiv.java.lang;
import edu.umd.cs.findbugs.apiv.VerifyResult;
import java.io.UnsupportedEncodingException;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import java.util.IllegalFormatException;
import edu.umd.cs.findbugs.apiv.java.util.regex.PatternV;

/**
 * Generic API invocation verifier for java.lag.String.
 * Return ARGUMENT_ERROR if an argument is out of bounds
 *
 * @see edu.umd.cs.findbugs.detect.ApiVerify
 * @author  Diomidis Spinellis
 * @version $Revision: 1.1.1.1 $
 */
public class StringV {

    public VerifyResult.Value String(boolean isConstant[], String t, char[] value, int offset, int count) {
	VerifyResult.Value r = VerifyResult.Value.CORRECT;

	if (isConstant[2]) {
	    if (offset < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[1]) {
		if (offset >= value.length)
		    return VerifyResult.Value.ARGUMENT_ERROR;
	    } else
		r = VerifyResult.Value.PART_VERIFIED;
	}
	if (isConstant[3]) {
	    if (count < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[1]) {
		if (count >= value.length)
		    return VerifyResult.Value.ARGUMENT_ERROR;
	    } else
		r = VerifyResult.Value.PART_VERIFIED;
	}
	return r;
    }

    public VerifyResult.Value String(boolean isConstant[], String t, int[] codePoints, int offset, int count) {
	VerifyResult.Value r = VerifyResult.Value.CORRECT;

	if (isConstant[2]) {
	    if (offset < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[1]) {
		if (offset >= codePoints.length)
		    return VerifyResult.Value.ARGUMENT_ERROR;
	    } else
		r = VerifyResult.Value.PART_VERIFIED;
	}
	if (isConstant[3]) {
	    if (count < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[1]) {
		if (count >= codePoints.length)
		    return VerifyResult.Value.ARGUMENT_ERROR;
	    } else
		r = VerifyResult.Value.PART_VERIFIED;
	}
	if (isConstant[1]) {
	    try {
		String s = new String(codePoints, 0, codePoints.length);
	    } catch (IllegalArgumentException e) {
		return VerifyResult.Value.ARGUMENT_ERROR;
	    }
	} else
	    r = VerifyResult.Value.PART_VERIFIED;
	return r;
    }

    public VerifyResult.Value String(boolean isConstant[], String t, byte[] bytes, int offset, int length, String charsetName) {
	VerifyResult.Value r = VerifyResult.Value.CORRECT;

	/*
	 * Assume that the charsets available here are the same as those
	 * available in the actual execution environment.
	 */
	if (isConstant[4]) {
	    try {
		String x = new String(new byte[]{}, charsetName);
	    } catch (UnsupportedEncodingException e) {
		return VerifyResult.Value.NAME_ERROR;
	    }
	} else
	    r = VerifyResult.Value.PART_VERIFIED;

	if (isConstant[2]) {
	    if (offset < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[1]) {
		if (offset >= bytes.length)
		    return VerifyResult.Value.ARGUMENT_ERROR;
	    } else
		r = VerifyResult.Value.PART_VERIFIED;
	}

	if (isConstant[3]) {
	    if (length < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[1]) {
		if (length >= bytes.length)
		    return VerifyResult.Value.ARGUMENT_ERROR;
	    } else
		r = VerifyResult.Value.PART_VERIFIED;
	}
	return r;
    }

    public VerifyResult.Value String(boolean isConstant[], String t, byte[] bytes, int offset, int length) {
	VerifyResult.Value r = VerifyResult.Value.CORRECT;

	if (isConstant[2]) {
	    if (offset < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[1]) {
		if (offset >= bytes.length)
		    return VerifyResult.Value.ARGUMENT_ERROR;
	    } else
		r = VerifyResult.Value.PART_VERIFIED;
	}

	if (isConstant[3]) {
	    if (length < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[1]) {
		if (length >= bytes.length)
		    return VerifyResult.Value.ARGUMENT_ERROR;
	    } else
		r = VerifyResult.Value.PART_VERIFIED;
	}
	return r;
    }

    public VerifyResult.Value String(boolean isConstant[], String t, byte[] bytes, String charsetName) {
	/*
	 * Assume that the charsets available here are the same as those
	 * available in the actual execution environment.
	 */
	if (isConstant[2]) {
	    try {
		String x = new String(new byte[]{}, charsetName);
	    } catch (UnsupportedEncodingException e) {
		return VerifyResult.Value.NAME_ERROR;
	    }
	    return VerifyResult.Value.CORRECT;
	} else
	    return VerifyResult.Value.UNVERIFIED;
    }

    public VerifyResult.Value charAt(boolean isConstant[], String t, int index) {
	if (isConstant[1]) {
	    if (index < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[0] && index >= t.length())
		return VerifyResult.Value.ARGUMENT_ERROR;
	}
	return VerifyResult.qualifiedCorrect(isConstant);
    }

    public VerifyResult.Value codePointAt(boolean isConstant[], String t, int index) {
	return charAt(isConstant, t, index);
    }

    public VerifyResult.Value codePointBefore(boolean isConstant[], String t, int index) {
	if (isConstant[1]) {
	    if (index < 1)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[0] && index > t.length())
		return VerifyResult.Value.ARGUMENT_ERROR;
	}
	return VerifyResult.qualifiedCorrect(isConstant);
    }

    public VerifyResult.Value codePointCount(boolean isConstant[], String t, int beginIndex, int endIndex) {
	if (isConstant[1]) {
	    if (beginIndex < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[2])
		if (beginIndex > endIndex)
		    return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[0] && beginIndex > t.length())
		return VerifyResult.Value.ARGUMENT_ERROR;
	}
	if (isConstant[0] && isConstant[2] && endIndex > t.length())
	    return VerifyResult.Value.ARGUMENT_ERROR;
	return VerifyResult.qualifiedCorrect(isConstant);
    }

    public VerifyResult.Value offsetByCodePoints(boolean isConstant[], String t, int index, int codePointOffset) {
	if (isConstant[1]) {
	    if (index < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[0] && index > t.length())
		return VerifyResult.Value.ARGUMENT_ERROR;
	}
	if (isConstant[0] && isConstant[1] && isConstant[2]) {
	    try {
		int x = t.offsetByCodePoints(index, codePointOffset);
		return VerifyResult.Value.CORRECT;
	    } catch (IndexOutOfBoundsException e) {
		return VerifyResult.Value.ARGUMENT_ERROR;
	    }
	}
	return VerifyResult.qualifiedCorrect(isConstant);
    }

    public VerifyResult.Value getChars(boolean isConstant[], String t, int srcBegin, int srcEnd, char[] dst, int dstBegin) {
	if (isConstant[1] && srcBegin < 0)
	    return VerifyResult.Value.ARGUMENT_ERROR;
	if (isConstant[1] && isConstant[2] && srcBegin > srcEnd)
	    return VerifyResult.Value.ARGUMENT_ERROR;
	if (isConstant[0] && isConstant[2] && srcEnd > t.length())
	    return VerifyResult.Value.ARGUMENT_ERROR;
	if (isConstant[4] && dstBegin < 0)
	    return VerifyResult.Value.ARGUMENT_ERROR;
	if (isConstant[1] && isConstant[2] && isConstant[3] && isConstant[4] &&  dstBegin + (srcEnd - srcBegin) > dst.length)
	    return VerifyResult.Value.ARGUMENT_ERROR;
	return VerifyResult.qualifiedCorrect(isConstant);
    }

    public VerifyResult.Value getBytes(boolean isConstant[], String t, String charsetName) {
	/*
	 * Assume that the charsets available here are the same as those
	 * available in the actual execution environment.
	 */
	if (isConstant[1]) {
	    try {
		byte[] b = "x".getBytes(charsetName);
	    } catch (UnsupportedEncodingException e) {
		return VerifyResult.Value.NAME_ERROR;
	    }
	    return VerifyResult.Value.CORRECT;
	} else
	    return VerifyResult.Value.UNVERIFIED;
    }

    public VerifyResult.Value contentEquals(boolean isConstant[], String t, StringBuffer sb) {
	if (isConstant[1]) {
	    if (sb == null)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    else
		return VerifyResult.Value.CORRECT;
	} else
	    return VerifyResult.Value.UNVERIFIED;
    }

    public VerifyResult.Value contentEquals(boolean isConstant[], String t, CharSequence sb) {
	if (isConstant[1]) {
	    if (sb == null)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    else
		return VerifyResult.Value.CORRECT;
	} else
	    return VerifyResult.Value.UNVERIFIED;
    }

    public VerifyResult.Value substring(boolean isConstant[], String t, int beginIndex) {
	if (isConstant[1]) {
	    if (beginIndex < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[0])
		if (beginIndex > t.length())
		    return VerifyResult.Value.ARGUMENT_ERROR;
		else
		    return VerifyResult.Value.CORRECT;
	    else
		return VerifyResult.Value.ARG_VERIFIED;
	} else
	    return VerifyResult.Value.UNVERIFIED;
    }

    public VerifyResult.Value substring(boolean isConstant[], String t, int beginIndex, int endIndex) {
	if (isConstant[1]) {
	    if (beginIndex < 0)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[2])
		if (beginIndex > endIndex)
		    return VerifyResult.Value.ARGUMENT_ERROR;
	    if (isConstant[0])
		if (beginIndex > t.length())
		    return VerifyResult.Value.ARGUMENT_ERROR;
	}
	if (isConstant[0] && isConstant[2] && endIndex > t.length())
	    return VerifyResult.Value.ARGUMENT_ERROR;
	return VerifyResult.qualifiedCorrect(isConstant);
    }

    public VerifyResult.Value subSequence(boolean isConstant[], String t, int beginIndex, int endIndex) {
	return substring(isConstant, t, beginIndex, endIndex);
    }

    public VerifyResult.Value matches(boolean isConstant[], String t, String regex) {
	boolean ic[] = new boolean[3];
	ic[0] = false;
	ic[1] = isConstant[2];
	ic[2] = isConstant[1];
	return PatternV.matches(ic, null, regex, t);
    }

    public VerifyResult.Value contains(boolean isConstant[], String t, CharSequence s) {
	if (isConstant[1])
	    if (s == null)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    else
		return VerifyResult.Value.CORRECT;
	else
	    return VerifyResult.Value.UNVERIFIED;
    }

    public VerifyResult.Value replaceFirst(boolean isConstant[], String t, String regex, String replacement) {
	if (isConstant[1] && regex == null)
		return VerifyResult.Value.ARGUMENT_ERROR;
	if (isConstant[2] && replacement == null)
	    return VerifyResult.Value.ARGUMENT_ERROR;
	if (isConstant[1]) {
	    try {
		Pattern p = Pattern.compile(regex);
		return VerifyResult.qualifiedCorrect(isConstant);
	    } catch (PatternSyntaxException e) {
		return VerifyResult.Value.SYNTAX_ERROR;
	    }
	} else
		return VerifyResult.Value.UNVERIFIED;
    }

    public VerifyResult.Value replaceAll(boolean isConstant[], String t, String regex, String replacement) {
	return replaceFirst(isConstant, t, regex, replacement);
    }

    public VerifyResult.Value split(boolean isConstant[], String t, String regex, int limit) {
	if (isConstant[1]) {
	    if (regex == null)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    try {
		Pattern p = Pattern.compile(regex);
		return VerifyResult.Value.CORRECT;
	    } catch (PatternSyntaxException e) {
		return VerifyResult.Value.SYNTAX_ERROR;
	    }
	} else
		return VerifyResult.Value.UNVERIFIED;
    }

    public VerifyResult.Value split(boolean isConstant[], String t, String regex) {
	return split(isConstant, t, regex, 0);
    }

    public static VerifyResult.Value format(boolean isConstant[], String t, String format, Object... args) {
	if (isConstant[1]) {
	    if (format == null)
		return VerifyResult.Value.ARGUMENT_ERROR;
	    try {
		String s = java.lang.String.format(format, args);
		return VerifyResult.Value.CORRECT;
	    } catch (IllegalFormatException e) {
		return VerifyResult.Value.SYNTAX_ERROR;
	    }
	} else
		return VerifyResult.Value.UNVERIFIED;
    }

    public static VerifyResult.Value valueOff(boolean isConstant[], String t, int offset, int count) {
	if (isConstant[1] && offset < 0)
	    return VerifyResult.Value.ARGUMENT_ERROR;
	if (isConstant[2] && count < 0)
	    return VerifyResult.Value.ARGUMENT_ERROR;
	return VerifyResult.qualifiedCorrect(isConstant);
    }

    public static VerifyResult.Value copyValueOff(boolean isConstant[], String t, int offset, int count) {
	return valueOff(isConstant, t, offset, count);
    }

}
